/*!
 * @edoc/resolver-qa 
 * (c) 2020-2021 Alex
 * Released under the MIT License.
 */
'use strict';

var javascriptStringify = require('javascript-stringify');
var chalk = require('chalk');
var pathFilter = require('@esydoc/path-filter');
var crypto = require('crypto');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var crypto__default = /*#__PURE__*/_interopDefaultLegacy(crypto);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __spreadArrays() {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
}

var CodeNode = /** @class */ (function () {
    function CodeNode() {
    }
    return CodeNode;
}());
var CodeGenerator = /** @class */ (function () {
    function CodeGenerator() {
    }
    return CodeGenerator;
}());
var EventResolver = /** @class */ (function () {
    function EventResolver() {
    }
    EventResolver.prototype.emitCode = function () {
        return this.generator.generate();
    };
    return EventResolver;
}());

// 负责作用域收集变量，并安全地使用
var Context = /** @class */ (function () {
    function Context() {
        this.scopeSet = new Set();
    }
    Context.prototype.wirte = function (variableName) {
        this.scopeSet.add(variableName);
    };
    Context.prototype.read = function (variableName) {
        if (this.scopeSet.has(variableName))
            return variableName;
        throw new Error("can't found the var name '" + variableName + ", are you sure decarle it before this code is runed?'");
    };
    return Context;
}());

var GlobalContext = new Context();

var hash = 0;
function genTryGetFnCode() {
    GlobalContext.wirte('tryGetFn');
    return "\n    function tryGetFn(fnGetter) {\n      try {\n        const fn = fnGetter()\n        if (typeof fn !== \"function\") {\n          throw new Error('skiped - fn is not a function')\n        } else {\n          return fn\n        }\n      } catch(err) {\n        throw new Error('skiped - not found fn')\n      }\n    }\n  ";
}
function genResolveFnExcuteResultCode() {
    GlobalContext.read('assert');
    GlobalContext.wirte('resolveFnExcuteResult');
    return "\n    function resolveFnExcuteResult(excutor, excutorWrapper, args, expectValue, done) {\n      let res\n\n      // excutorWrapper mark api error that occured itself in sync or async\n      try {\n        res = excutor.apply(null, args)\n        if (res instanceof Promise) {\n          res.then((result) => {\n            excutorWrapper.excutorReturns.push(result)\n            try {\n              assert(result, expectValue)\n              done()\n            } catch(e) {\n              done(e)\n            }\n          }).catch((e) => {\n            excutorWrapper.excutorReturns.push(e)\n            done(e)\n          })\n        } else {\n          excutorWrapper.excutorReturns.push(res)\n          assert(res, expectValue) // mocha will catch error in sync, we no need to do it.\n          done()\n        }\n      } catch(e) {\n        excutorWrapper.excutorReturns.push(e)\n        done(e)\n      }\n    }\n  ";
}
var ApiCallNode = /** @class */ (function (_super) {
    __extends(ApiCallNode, _super);
    function ApiCallNode(path, options) {
        var _this = _super.call(this) || this;
        _this.id = hash++;
        _this.path = path;
        _this.pathPrefix = options.pathPrefix;
        _this.deps = options.deps || [];
        var name = path.split('.').pop() || 'unknow';
        _this.name = name + '_' + _this.id;
        return _this;
    }
    Object.defineProperty(ApiCallNode.prototype, "fullPath", {
        get: function () {
            return this.pathPrefix + "." + this.path;
        },
        enumerable: false,
        configurable: true
    });
    ApiCallNode.prototype.generate = function () {
        return "\n      // created by ApiCallNode\n      (function create_" + this.name + "_WrapperFn () {\n        let rawFn\n        try {\n          rawFn = " + GlobalContext.read('tryGetFn') + "(() => " + this.fullPath + ")\n        } catch(err) {\n          rawFn = function deliverError() {\n            throw err\n          }\n        }\n      \n        const spyFn = " + GlobalContext.read('sinon') + ".spy(rawFn)\n        const wrapperFn = function _" + this.name + "_Wrapper(args, expectValue, next) {\n          wrapperFn.expects.push(" + GlobalContext.read('helper') + ".v2type(expectValue))\n          " + GlobalContext.read('resolveFnExcuteResult') + "(spyFn, wrapperFn, args, expectValue, next)\n        }\n        \n        // expose some api to outside\n        wrapperFn.excutor = spyFn\n        wrapperFn.deps = " + JSON.stringify(this.deps) + "\n        wrapperFn.expects = [] // \u653E\u7F6E\u7684\u662F\u671F\u671B\u7C7B\u578B\n        wrapperFn.excutorReturns = []\n        wrapperFn.id = '" + ('_' + this.name + '_Wrapper') + "'\n\n        return wrapperFn\n      })()\n    ";
    };
    ApiCallNode.genTryGetFnCode = genTryGetFnCode;
    ApiCallNode.genResolveFnExcuteResultCode = genResolveFnExcuteResultCode;
    return ApiCallNode;
}(CodeNode));

var KeyValueNode = /** @class */ (function (_super) {
    __extends(KeyValueNode, _super);
    function KeyValueNode(key, value) {
        var _this = _super.call(this) || this;
        _this.key = key;
        _this.value = value;
        return _this;
    }
    return KeyValueNode;
}(CodeNode));
var MarkItemNode = /** @class */ (function (_super) {
    __extends(MarkItemNode, _super);
    function MarkItemNode() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    MarkItemNode.genMarkObjectCode = function (nodeQueue) {
        var keyValueString = function (nodeQueue) {
            return nodeQueue.map(function (node) {
                return "\"" + node.key + "\": " + node.generate();
            });
        };
        return "{\n      " + keyValueString(nodeQueue).join(',\n') + "\n    }";
    };
    MarkItemNode.prototype.generate = function () {
        return "\n      {\n        key: \"" + this.key + "\",\n        value: " + this.value + "\n      }\n    ";
    };
    return MarkItemNode;
}(KeyValueNode));
var ObjectItemNode = /** @class */ (function (_super) {
    __extends(ObjectItemNode, _super);
    function ObjectItemNode() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    ObjectItemNode.genObjectCode = function (nodeQueue) {
        var keyValueString = function (nodeQueue) {
            return nodeQueue.map(function (node) {
                return node.generate();
            });
        };
        return "{\n      " + keyValueString(nodeQueue).join(',\n') + "\n    }";
    };
    ObjectItemNode.prototype.generate = function () {
        return "\n      " + this.key + ": " + this.value + "\n    ";
    };
    return ObjectItemNode;
}(KeyValueNode));

var GetValueNode = /** @class */ (function (_super) {
    __extends(GetValueNode, _super);
    function GetValueNode(options) {
        var _this = _super.call(this) || this;
        _this.options = options;
        _this.isFn = typeof options.perperty === 'function';
        _this.perperty = GetValueNode.toString(options.perperty);
        return _this;
    }
    GetValueNode.toString = function (args) {
        return typeof args === 'function' ? args.toString() : javascriptStringify.stringify(args);
    };
    GetValueNode.prototype.generate = function (args) {
        var perperty = this.perperty;
        if (this.isFn) {
            return "(" + perperty + ")(" + (args != '' ? "..." + args : '') + ")";
        }
        return perperty;
    };
    return GetValueNode;
}(CodeNode));

function genTransfromDepsCode() {
    GlobalContext.wirte('transfromDeps');
    return "\n    function transfromDeps(deps = [], cb) {\n      return deps.map((apiName) => {\n        const apiConfig = apiMark[apiName]\n        if (!apiConfig) throw new Error('The api name of ' + apiName + ' was not found in apiMark')\n        return cb(apiConfig.value, apiName)\n      })\n    }\n  ";
}
var TestCaseNode = /** @class */ (function (_super) {
    __extends(TestCaseNode, _super);
    function TestCaseNode(options) {
        var _this = _super.call(this) || this;
        _this.options = options;
        _this.args = new GetValueNode({ perperty: options.args });
        _this.expect = new GetValueNode({ perperty: options.expect });
        return _this;
    }
    TestCaseNode.prototype.genDoneByUserControlFn = function (doneKey) {
        var onDone = this.options.onDone;
        if (!onDone)
            return doneKey;
        return "\n    function doneByUserControl(e) {\n      " + new GetValueNode({ perperty: onDone }).generate("[e, " + doneKey + "]") + "\n    }\n    ";
    };
    TestCaseNode.prototype.useTryCatch = function (runCode, handeErrorCode) {
        return "\n      try {\n        " + runCode + "\n      } catch(e) {\n        " + handeErrorCode('e') + "\n      }\n    ";
    };
    TestCaseNode.prototype.generate = function () {
        var _a = this.options, apiPath = _a.apiPath, sort = _a.sort, categroySort = _a.categroySort, _b = _a.deps, deps = _b === void 0 ? [] : _b;
        var shouldExportDeps = this.args.isFn || this.expect.isFn;
        var depsDeclareCode = shouldExportDeps ? 'let deps = [] \r\n' : '';
        var depsAssginCode = shouldExportDeps ? 'deps =' : '';
        return "\n      it('No." + sort + " $[" + apiPath + "] - \u7B2C" + categroySort + "\u4E2A\u6D4B\u8BD5\u7528\u4F8B', (done) => {\n        const testRunner = " + GlobalContext.read('apiMark') + "['" + apiPath + "'].value\n        const doneCb = " + this.genDoneByUserControlFn('done') + "\n\n        " + (deps.length
            ? depsDeclareCode +
                this.useTryCatch(depsAssginCode +
                    (GlobalContext.read('transfromDeps') + "(" + GetValueNode.toString(deps) + ", (thisTestRunner, apiName) => {\n              const lastReturn = thisTestRunner.excutorReturns[thisTestRunner.excutorReturns.length - 1]\n              if (lastReturn instanceof Error) {\n                throw new Error('The ' + apiName + ' ocurred an error, it depend on " + apiPath + ".')\n              }\n              return lastReturn\n            })"), function (errKey) { return "return doneCb(" + errKey + ")"; })
            : '') + "\n\n        const args = " + this.args.generate(deps.length ? 'deps' : '') + "\n        const expectValue = " + this.expect.generate(deps.length ? 'deps' : '') + "\n        \n        testRunner(args, expectValue, doneCb)\n      })\n    ";
    };
    TestCaseNode.genTransfromDepsCode = genTransfromDepsCode;
    return TestCaseNode;
}(CodeNode));

var genInitTestEnvCode = function () {
    GlobalContext.wirte('helper');
    GlobalContext.wirte('sinon');
    return "\n    import helper from './QA-handle/helper'\n    import sinon from 'sinon'\n  ";
};
// export const isMatch = (a: any, b: any) => {
//   if (a === b) return true
//   if (Object.is(a, b)) return true
//   if (isLooseEqual(a, b)) return true
//   return false
// }
// const isObject = (v: any) => {
//   return typeof v === 'object' && v !== null
// }
// // to value type
// export const isLooseEqual = (a: any, b: any) => {
//   if (Array.isArray(a) && Array.isArray(b)) {
//     if (a.length !== b.length) {
//       return false
//     }
//     let i = 0,
//       l = a.length
//     while (i < l) {
//       const result = isLooseEqual(a[i], b[i])
//       if (!result) return false
//       i++
//     }
//     return true
//   } else if (isObject(a) && isObject(b)) {
//     if (Object.keys(a).length !== Object.keys(b).length) {
//       return false
//     }
//     for (const key in a) {
//       const result = isLooseEqual(a[key], b[key])
//       if (!result) return false
//     }
//     return true
//   } else {
//     return a === b
//   }
// }
var genAssertFn = function () {
    GlobalContext.wirte('assert');
    return "\n    // \u4EE5a\u4F5C\u4E3A\u53C2\u8003\u70B9\uFF0Cb\u7684\u957F\u5EA6\u4E0D\u80FD\u5C0F\u4E8Ea\n    const isSameValueType = (a, b) => {\n      const aPath = []\n      const bPath = []\n\n      const isObject = (v) => {\n        return typeof v === \"object\" && v !== null\n      }\n    \n      const getValueType = (v) => {\n        if (Array.isArray(v)) return 'array'\n        if (isObject(v)) return 'object'\n        if (v === null) return null\n        return typeof v\n      }\n    \n      const check = (a, b) => {\n        if (Array.isArray(a) && Array.isArray(b)) {\n          if (a.length > b.length) {\n            return false\n          }\n      \n          let i = 0, l = a.length\n          while (i < l) {\n            const result = check(a[i], b[i])\n            if (!result) {\n              aPath.push(\"index:\" + i + \",\" + \"value:\" + getValueType(a[i]))\n              bPath.push(\"index:\" + i + \",\" + \"value:\" + getValueType(b[i]))\n              return false\n            }\n    \n            i++\n          }\n      \n          return true\n        } else if (isObject(a) && isObject(b)) {\n          if (Object.keys(a).length > Object.keys(b).length) {\n            aPath.push(\"key length:\" + a.length)\n            bPath.push(\"key length:\" + b.length)\n            return false\n          }\n      \n          for (const key in a) {\n            const result = check(a[key], b[key])\n    \n            if (!result) {\n              aPath.push(\"key:\" + key + \",\" + \"value:\" + getValueType(a[key]))\n              bPath.push(\"key:\" + key + \",\" + \"value:\" + getValueType(b[key]))\n              return false\n            }\n          }\n      \n          return true\n        } else {\n          return typeof a === typeof b\n        }\n      }\n    \n      const res = check(a, b)\n      return [res, aPath.reverse(), bPath.reverse()]\n    }\n\n    const assert = (returnValue, expectValue) => {\n      // \u5DF2\u671F\u671B\u503C\u4E3A\u57FA\u51C6\n      const [pass, aPath, bPath ] = isSameValueType(expectValue, returnValue)\n      if (!pass) {\n        let errMsg = \"\u65AD\u8A00\u5931\u8D25\uFF0C\u5B9E\u9645\u503C\u4E0E\u671F\u671B\u503C\u7684\u7C7B\u578B\u4E0D\u5339\u914D\" + '\\n'\n        \n        if (aPath.length) {\n          errMsg += '\u671F\u671B\uFF1A' +JSON.stringify(aPath, null, 2) + '\\n' + '\u5B9E\u9645\uFF1A' + JSON.stringify(bPath, null, 2)\n        }\n    \n        throw new Error(errMsg)\n      }\n    }\n  ";
};

var createNode = function (value) {
    return {
        value: value,
        nextSibling: null,
        child: null,
        return: null,
    };
};
var toBTree = function (rawTree) {
    var root = createNode(rawTree.value);
    var current;
    var last;
    if (rawTree.deps) {
        rawTree.deps.forEach(function (subValue, i) {
            if (i === 0) {
                current = last = root.child = createNode(subValue);
            }
            else {
                current = createNode(subValue);
                last.nextSibling = current;
                last = current;
            }
            current.return = root;
        });
    }
    return root;
};
// 优先深度遍历-反转
function beginWork(root, cb) {
    var stack = [];
    var current = root;
    while (current || stack.length) {
        if (typeof current !== 'undefined') {
            if (cb(current))
                return current;
            current.child && stack.push(current.child);
            current.nextSibling && stack.push(current.nextSibling);
        }
        current = stack.pop();
    }
    return null;
}
var traverseParentPathCheckCircularRef = function (node) {
    var val = node.value;
    var parent = node.return;
    while (parent !== null) {
        // 节点往上遍历，检查是否有循环依赖
        if (parent.value === val) {
            throw new Error("the value " + val + " be circulared reference by parent path");
        }
        parent = parent.return;
    }
};
var combineNode = function (node, hostRoot) {
    if (hostRoot.child !== null) {
        if (node.child === null) {
            node.child = hostRoot.child;
            // node.child 是一颗斜树
            beginWork(node.child, function (internalNode) {
                internalNode.return = node; // 保持新父节点引用
                // 检查循环依赖
                traverseParentPathCheckCircularRef(internalNode);
            });
        }
    }
    else if (hostRoot.nextSibling !== null) { // hostRoot压根就没有nextSibling，这段代码为了以后
        if (node.nextSibling === null) {
            node.nextSibling = hostRoot.nextSibling;
            beginWork(node.nextSibling, function (internalNode) {
                internalNode.return = node.return; // 保持新父节点引用
                // 检查循环依赖
                traverseParentPathCheckCircularRef(internalNode);
            });
        }
    }
};
var combineForest2BTree = function (forest) {
    var rootTree = forest.shift();
    (function forest2tree() {
        if (!forest.length)
            return;
        var hostRoot = forest.shift();
        if (hostRoot && rootTree) {
            var matchNode = beginWork(rootTree, function (node) {
                return node.value === hostRoot.value;
            });
            if (matchNode !== null) {
                combineNode(matchNode, hostRoot);
            }
            else {
                // rootTree增加一个兄弟节点
                var current = rootTree;
                while (current !== null) {
                    if (current.nextSibling !== null) {
                        current = current.nextSibling;
                    }
                    else {
                        current.nextSibling = hostRoot;
                        break;
                    }
                }
            }
        }
        forest2tree();
    })();
    return rootTree;
};
function shrink(path) {
    var temp = {};
    path.forEach(function (val) {
        temp[val] = 1;
    });
    return Object.keys(temp);
}
function analysisDeps(rawForest) {
    if (rawForest.length === 0)
        return [];
    var depPath = [];
    var bForest = rawForest.map(toBTree);
    var fullTree = combineForest2BTree(bForest);
    if (fullTree) {
        beginWork(fullTree, function (node) {
            depPath.push(node.value);
        });
    }
    return shrink(depPath.reverse());
}

function genShouldExistTaseCaseFnCode() {
    GlobalContext.wirte('shouldExistTaseCase');
    return "\n    function shouldExistTaseCase (targetPath, apiQueue) {\n      // \u5148\u68C0\u67E5\u5B83\u662F\u4E0D\u662F\u81EA\u8EAB\n      if (apiQueue.find((api) => targetPath === api.path)) {\n        return true\n      }\n    \n      // \u518D\u68C0\u67E5\u5B83\u662F\u4E0D\u662F\u4F9D\u8D56\n      const depsArray = apiQueue.reduce((acc, item) => {\n        acc = acc.concat(item.ref.deps)\n        return acc\n      }, [])\n    \n      return depsArray.find((dep) => dep === targetPath)\n    }\n  ";
}
function outputTestCaseCode(options) {
    var testCaseNode = new TestCaseNode(options);
    return GlobalContext.read('shouldExistTaseCase') + "('" + options.apiPath + "', " + GlobalContext.read('apiQueue') + ") && " + testCaseNode.generate();
}
var TestCaseNum = 0;
var ApiQACodeGenerator = /** @class */ (function (_super) {
    __extends(ApiQACodeGenerator, _super);
    function ApiQACodeGenerator(options) {
        var _this = _super.call(this) || this;
        _this.nodeConfigMap = new Map();
        _this.options = options;
        return _this;
    }
    ApiQACodeGenerator.prototype.collectNode = function (config) {
        this.nodeConfigMap.set(config.node.path, config);
    };
    ApiQACodeGenerator.prototype.ouputNewNodeConfigQueueByDeps = function () {
        console.log(chalk__default['default'].magenta('Begining analyze Dependencies...'));
        var rawForest = [];
        var nodeConfigMap = this.nodeConfigMap;
        nodeConfigMap.forEach(function (config) {
            rawForest.push({
                value: config.node.path,
                deps: config.deps,
            });
        });
        var flatenTree = analysisDeps(rawForest);
        var sortConigQueue = [];
        flatenTree.forEach(function (path) {
            var config = nodeConfigMap.get(path);
            config && sortConigQueue.push(config);
        });
        return sortConigQueue;
    };
    ApiQACodeGenerator.prototype.getHead = function () {
        var markItems = [];
        this.nodeConfigMap.forEach(function (config) {
            markItems.push(new MarkItemNode(config.node.path, config.node.generate()));
        });
        GlobalContext.wirte('apiMark');
        GlobalContext.wirte('hostContext');
        return "\n      export const apiMark = " + MarkItemNode.genMarkObjectCode(markItems) + "\n      export const hostContext = " + GetValueNode.toString(this.options.hostContext || {}) + "\n    ";
    };
    ApiQACodeGenerator.prototype.getBody = function () {
        // gen test case by ApiCallNode
        return this.ouputNewNodeConfigQueueByDeps()
            .map(function (config) {
            var node = config.node, expectQueue = config.expectQueue, args2d = config.args2d, deps = config.deps, onDone = config.onDone;
            // 兼容一维数组
            if (typeof args2d === 'function' || // args2d is () => {}
                !args2d.length || // args2d is []
                (args2d.find(function (item) { return !Array.isArray(item); }) &&
                    !Array.isArray(expectQueue)) // args2d is [{}, number, string, ...]
            ) {
                return [
                    outputTestCaseCode({
                        apiPath: node.path,
                        args: args2d,
                        expect: expectQueue,
                        sort: ++TestCaseNum,
                        categroySort: 1,
                        deps: deps,
                        onDone: onDone
                    }),
                ];
            }
            else {
                // 二维数组
                return args2d.map(function (args, j) {
                    return outputTestCaseCode({
                        apiPath: node.path,
                        args: args,
                        expect: expectQueue[j],
                        sort: ++TestCaseNum,
                        categroySort: j + 1,
                        deps: deps,
                        onDone: onDone
                    });
                });
            }
        })
            .reduce(function (acc, item) {
            acc = acc.concat(item);
            return acc;
        }, [])
            .join('\n');
    };
    ApiQACodeGenerator.prototype.getFoot = function () {
        return '';
    };
    ApiQACodeGenerator.prototype.generate = function () {
        if (this.nodeConfigMap.size === 0)
            return '';
        GlobalContext.wirte('apiQueue');
        return "\n      " + genInitTestEnvCode() + "\n      " + genAssertFn() + "\n\n      " + ApiCallNode.genTryGetFnCode() + "\n      " + ApiCallNode.genResolveFnExcuteResultCode() + "\n      " + TestCaseNode.genTransfromDepsCode() + "\n      " + ApiQACodeGenerator.genShouldExistTaseCaseFnCode() + "\n\n      " + this.getHead() + "\n\n      export function injectTestSuite (apiQueue) {\n        describe('QA testing', () => {\n          " + this.getBody() + "\n        })\n      }\n\n      " + this.getFoot() + "\n    ";
    };
    ApiQACodeGenerator.genShouldExistTaseCaseFnCode = genShouldExistTaseCaseFnCode;
    return ApiQACodeGenerator;
}(CodeGenerator));

var resolveTag = function (tagName) {
    var filterWrapperRE = /([a-zA-Z]+)\.?<(.*)>/i;
    var match = tagName.match(filterWrapperRE);
    if (match == null)
        return { tagName: tagName, isNested: false };
    var parentTagName = match[1];
    var subTagName = match[2];
    if (parentTagName === 'Promise') {
        return resolveTag(subTagName);
    }
    return { tagName: parentTagName, subTagName: subTagName, isNested: true };
};
var DATA_NODE_ID_PREFIX = '__data_node__';
var createNode$1 = function (name, type, properties) {
    var resolvedTag = resolveTag(type.names[0]);
    var id = DATA_NODE_ID_PREFIX + crypto__default['default'].randomBytes(6).toString('hex');
    if (!resolvedTag.isNested) {
        return {
            name: name,
            properties: properties,
            type: resolvedTag.tagName,
            id: id,
        };
    }
    // for example: Array<xxx>
    var subTagName = resolvedTag.subTagName;
    var tagName = resolvedTag.tagName;
    return {
        id: id,
        name: tagName,
        type: tagName,
        properties: [
            {
                name: subTagName,
                type: {
                    names: [subTagName],
                },
            },
        ],
    };
};
var record = function (store, v, name) {
    if (Array.isArray(store)) {
        store.push(v);
    }
    else {
        store[name] = v;
    }
};
var transform = function (rawNode, dataMap, store) {
    var node = createNode$1(rawNode.name, rawNode.type, rawNode.properties || rawNode.params // 统一抽象为一个properties
    );
    var name = node.name;
    var type = node.type;
    switch (type) {
        case 'Object':
        case 'Array':
            var initValue_1 = type === 'Object' ? {} : [];
            record(store, initValue_1, name);
            if (node.properties) {
                node.properties.forEach(function (child) {
                    transform(child, dataMap, initValue_1);
                });
            }
            break;
        case 'number':
        case 'string':
        case 'Function':
        case 'function': // @callback type is function
        case 'undefined':
        case 'boolean':
        case 'null':
        case 'any':
            record(store, node, name);
            break;
        default:
            // 自定义结构体
            if (dataMap) {
                var subRawTree = dataMap.get(type);
                if (subRawTree) {
                    // 我们需要的是父节点上的name，而不是typeof定义的名字
                    name && (subRawTree.name = name);
                    transform(subRawTree, dataMap, store);
                }
                else {
                    console.log(chalk__default['default'].yellow("\u65E0\u6CD5\u5728dataMap\u4E2D\u627E\u5230type:" + type + "\u7684\u7ED3\u6784\u4F53"));
                }
            }
            else {
                console.log(chalk__default['default'].yellow("\u65E0\u6CD5\u89E3\u6790type:" + type + "\u7684\u7ED3\u6784\u4F53"));
            }
    }
};
var type2Value = function (v) {
    switch (v) {
        case 'number':
            return 0;
        case 'string':
            return '';
        case 'Function':
            return function () { };
        case 'null':
            return null;
        case 'boolean':
            return true;
        case 'undefined':
        case '*':
        case 'any':
            return undefined;
        default:
            throw new Error("\u65E0\u6CD5\u89E3\u6790\u503C\u7C7B\u578B\u4E3A{" + v + "}\u7684\u5B57\u6BB5");
    }
};
var traverseNode = function (v, cb) {
    function traverseObj(v) {
        if (Array.isArray(v)) {
            return v.map(function (v) {
                return traverseObj(v);
            });
        }
        else if (typeof v === 'object' && v !== null) {
            if (v.hasOwnProperty('id') &&
                typeof v.id === 'string' &&
                v.id.startsWith(DATA_NODE_ID_PREFIX)) {
                return cb(v); // 粒度掌控在nodeType上
            }
            return Object.keys(v).reduce(function (track, key) {
                var value = v[key];
                track[key] = traverseObj(value);
                return track;
            }, {});
        }
        else {
            return cb(v);
        }
    }
    return traverseObj(v);
};
var reflectRealValue = function (v) {
    return traverseNode(v, function (node) {
        if (node.type === 'function') {
            var properties = node.properties;
            if (!properties || !properties.length) {
                return new Function();
            }
            var args = properties.map(function (node) {
                return node.name || node.type.names[0];
            });
            // 我们需要动态构造函数
            return new (Function.bind.apply(Function, __spreadArrays([void 0], __spreadArrays(args, ['// do something']))))();
        }
        else {
            return type2Value(node.type);
        }
    });
};

function create(rawDataRoot, dataMap) {
    var reflectArgTypeTree = {};
    transform(rawDataRoot, dataMap, reflectArgTypeTree);
    return reflectRealValue(reflectArgTypeTree);
}
function createTmplConfig(content, key, nodeType) {
    var root = {
        name: 'root',
        properties: content,
        type: {
            names: [nodeType],
        },
    };
    var tmplConfig = {
        key: key,
        root: root,
    };
    return tmplConfig;
}
var outputMapping = {
    args: {
        reflectKey: 'params',
        format: function (code) {
            return "[" + code + "]"; // args是一个二维数组
        },
        nodeType: 'Array',
        defaultValue: '[[]]',
    },
    expect: {
        reflectKey: 'returns',
        format: function (code) {
            return "" + code; // expect是一个一维数组
        },
        nodeType: 'Array',
        defaultValue: '[]',
    },
};
var ConfigTemplate = /** @class */ (function () {
    function ConfigTemplate(fileDataMap) {
        this.fileDataMap = fileDataMap;
    }
    ConfigTemplate.createDefaultTemplate = function () {
        return ObjectItemNode.genObjectCode([
            new ObjectItemNode('args', '[[]]'),
            new ObjectItemNode('expect', '[]'),
            new ObjectItemNode('deps', '[]'),
        ]);
    };
    ConfigTemplate.prototype.create = function (rawDataRoot, filename, key, format) {
        var dataMap = this.fileDataMap.get(filename);
        var config = create(rawDataRoot, dataMap);
        var root = config.root;
        if (typeof root === 'object' && root !== null && Object.keys(root).length) {
            return new ObjectItemNode(key, format(javascriptStringify.stringify(root)));
        }
        return null;
    };
    ConfigTemplate.prototype.output = function (doclet) {
        var _this = this;
        var meta = doclet.meta;
        var configQueue = [];
        var shouldNotProcessKeys = ['deps'];
        var shouldProcessKeys = Object.keys(outputMapping);
        shouldProcessKeys.forEach(function (key) {
            // @ts-expect-error
            var outputConfig = outputMapping[key];
            var docletPropertyValue = doclet[outputConfig.reflectKey];
            if (docletPropertyValue) {
                var config = createTmplConfig(docletPropertyValue, key, outputConfig.nodeType);
                configQueue.push(config);
            }
        });
        if (configQueue.length) {
            var markNodes = configQueue
                .map(function (tmplConfig) {
                return _this.create(tmplConfig.root, meta.filename, tmplConfig.key, 
                // @ts-expect-error
                outputMapping[tmplConfig.key].format);
            })
                .filter(function (v) { return v !== null; });
            if (!markNodes.length) {
                return ConfigTemplate.createDefaultTemplate();
            }
            var existKeys_1 = __spreadArrays(shouldProcessKeys, shouldNotProcessKeys);
            markNodes.forEach(function (node, i) {
                var targetIndex;
                if ((targetIndex = existKeys_1.indexOf(node.key)) > -1) {
                    existKeys_1.splice(targetIndex, 1);
                }
            });
            var paddingNodes = existKeys_1.map(function (key) {
                // @ts-expect-error
                var outputConfig = outputMapping[key];
                var defaultValue = '[]';
                if (outputConfig) {
                    defaultValue = outputConfig.defaultValue;
                }
                return new ObjectItemNode(key, defaultValue);
            });
            return ObjectItemNode.genObjectCode(__spreadArrays(markNodes, paddingNodes));
        }
        return ConfigTemplate.createDefaultTemplate();
    };
    return ConfigTemplate;
}());

var ApiQACodeResolver = /** @class */ (function (_super) {
    __extends(ApiQACodeResolver, _super);
    function ApiQACodeResolver(scheduler, options) {
        var _this = _super.call(this) || this;
        _this.options = Object.assign({
            pathPrefix: '',
            verbose: true,
        }, options);
        _this.scheduler = scheduler;
        _this.configTemplate = new ConfigTemplate(scheduler.fileDataMap);
        _this.generator = new ApiQACodeGenerator({
            hostContext: _this.options.hostContext,
        });
        return _this;
    }
    ApiQACodeResolver.prototype.resolveApi = function (doclet, apiConfig) {
        var apiPathPrefix = this.options.qa.apiPathPrefix;
        var callPath = this.scheduler.getCallPath(doclet);
        var apiCallNode = new ApiCallNode(callPath, __assign(__assign({}, apiConfig), { pathPrefix: apiPathPrefix }));
        this.generator.collectNode({
            node: apiCallNode,
            args2d: apiConfig.args,
            expectQueue: apiConfig.expect,
            deps: apiConfig.deps,
            onDone: apiConfig.onDone,
        });
    };
    ApiQACodeResolver.prototype.resolveEvent = function (doclet, apiInfo) {
        var qa = this.options.qa;
        var configFilePathFragment = apiInfo.configFilePathFragment, configFilePath = apiInfo.configFilePath, apiConfig = apiInfo.apiConfig;
        var shouldRemain = pathFilter.assertRemain(configFilePathFragment, {
            context: this.options.context,
            excludes: qa.excludes,
            includes: qa.includes
        });
        if (shouldRemain) {
            console.log("\u6536\u96C6\u8282\u70B9\uFF1A" + chalk__default['default'].cyan(this.scheduler.getCallPath(doclet)));
            this.resolveApi(doclet, apiConfig);
        }
    };
    ApiQACodeResolver.prototype.template = function (doclet) {
        return this.configTemplate.output(doclet);
    };
    return ApiQACodeResolver;
}(EventResolver));
module.exports = ApiQACodeResolver;

module.exports = ApiQACodeResolver;
